;***************************************************************************;
;                          P R I N I T . A S M                              ;
;===========================================================================;
;       Keyboard TSR demonstration.                                         ;
;       Resets and initialises, or just resets (disables) printer.          ;
;       Assembled with Turbo Assembler version 3.0.                         ;
;       Linked with Turbo Linker version 5.1.                               ;
;---------------------------------------------------------------------------;
;       Copyright (c) Simon Groes, 1995                                     ;
;***************************************************************************;

        IDEAL
        DOSSEG
        MODEL   tiny
        LOCALS

;=====  EQUATES ================================================
ENVBLOCK        EQU     02Ch
;-----  BIOS VARIABLES
BIOS_DATASEG    EQU     0040h           ; Seg addr of BIOS data area.
BIOS_LPT1PORT   EQU     8h

;-----  SCAN CODES:
SC_R            EQU     13h
SC_Q            EQU     10h

;-----  KEY CODES:
CR              EQU     0Dh
LF              EQU     0Ah

;-----  KEY STATES:
KS_RSHIFT       EQU     00000001b
KS_LSHIFT       EQU     00000010b       ; KS = 'KeyState'
KS_CONTROL      EQU     00000100b
KS_ALT          EQU     00001000b       ; Alt key down.
KS_SCROLLOCK    EQU     00010000b       ; Scroll Lock on.
KS_NUMLOCK      EQU     00100000b       ; Num Lock on.
KS_CAPSLOCK     EQU     01000000b       ; Caps Lock on.
KS_INSERT       EQU     10000000b       ; Insert on.

;-----  PRINTER STATUS REGISTER BITS:
PS_NOTBUSY      EQU     10000000b       ; 1=not busy, 0=busy.
PS_ACK          EQU     01000000b       ; 1=acknowledge, 0=ready for next char.   
PS_NOPAPER      EQU     00100000b       ; 1=out of paper.
PS_ONLINE       EQU     00010000b       ; 1=printer online.
PS_ERROR        EQU     00001000b       ; 1=I/O error.
PS_TIMEOUT      EQU     00000001b       ; 1=Time-out.

;-----  PRINTER CONTROL REGISTER BITS:
PC_INIT         EQU     00000100b       ; Reset bit. 0=reset, 1=set.
PC_DATA         EQU     00000001b       ; 1=data transfer on.     
;---------------------------------------------------------------

        CODESEG

;=====  EXTERNAL ROUTINES ===========================================
;-----  From ARG module:
        EXTRN   ARG_ParseArgs:PROC, ARG_GetArgNumber:PROC, ARG_IsSwitch:PROC
;-----  From ERRCODE module:
        EXTRN   ERRCODE_GetErrMsg:PROC
;--------------------------------------------------------------------

        ORG     100h
;***************************************************************
;                     MAIN PROGRAM
;***************************************************************
Start:
        jmp     NEAR    Begin           ; Jump to position after TSR.

;********************************************************************
;                     TSR START
;********************************************************************

;===============================================================
; Interrupt:    =*=Int09=*=
;---------------------------------------------------------------
; Task:         Increment bKbhit variable at keyboard int (09h).
; Input:        none
; Output:       none
; Registers:    none
;===============================================================

;=====  VARIABLES ==============================================
szID09          db      'Prinit (c) Simon Helmut Groes',0  ; ID string.
fRunning        db      0               ; To avoid reentry.

LABEL   dwInt09LPtr     DWORD
wInt09Ofs       dw      0       ; Offset address of old 1Ch int.
wInt09Seg       dw      0       ; Segment address of old 1Ch int.
ID09            =       $ - szID09
SEG09           =       $ - wInt09Seg
OFS09           =       $ - wInt09Ofs
;---------------------------------------------------------------
PROC    Int09

        cmp     [BYTE cs:fDisk], 0      ; Is disk currently being accessed ?
        jne     @@Default               ; Yes: Execute only original int 9.
        cmp     [BYTE cs:fRunning], 0   ; Program already running ?
        jne     @@Default               ; Yes: Execute only original int 9.
        inc     [BYTE cs:fRunning]      ; No : Set [fRunning] to 1.
        
        pushf                           ; Simulate 'INT' instruction.
        call    [DWORD cs:dwInt09LPtr]  ; Call original int 9h.

;-----  New handler's processing.
        sti                             ; Allow interrupts.
        
        push    ax
        push    dx
        call    BIOS_GetKeyState        ; al = status bits. ah = 0.
        mov     dx, ax                  ; Store in dx.
        call    ReadKbPort              ; No char available if ax=0.
        or      al, al                  ; Scan code in al ?
        jz      @@Return                ; No : Return to original program.
        cmp     al, SC_R                ; al = 'R' scancode ?
        je      @@ControlKeys           ; Yes: Check control keys.
        cmp     al, SC_Q                ; al = 'Q' scancode ?
        jne     @@Return                ; No : Return to original program.

@@ControlKeys:
        test    dl, KS_CONTROL          ; Ctrl key down ?
        jz      @@Return                ; No : Return to original program.
        test    dl, KS_LSHIFT           ; Left Shift key down ?
        jnz     @@Activate              ; Yes: Process.
        test    dl, KS_RSHIFT           ; Right Shift key down ?
        jnz     @@Activate              ; Yes: Process.

@@Return:
;-----  Disable interrupts during cleanup and exit.
        cli                             ; Disable interrupts.
        pop     dx
        pop     ax
        dec     [BYTE cs:fRunning]      ; Reset [fRunning] flag to 0.
        
@@End:
        iret
@@Default:
        pushf                           ; Simulate 'INT' instruction.
        call    [DWORD cs:dwInt09LPtr]  ; Call original int 9h.
        iret

@@Activate:
        push    ax                      ; al=Scancode. Next func changes ax.
        call    BIOS_ReadChar           ; Empty buffer.
        pop     ax                      ; Restore ax.
        cmp     al, SC_R                ; Ctrl+Shift+R pressed ?
        je      @@Init                  ; Yes: Reset and initialise printer.
        cmp     al, SC_Q                ; Ctrl+Shift+Q pressed ?
        je      @@Reset                 ; Yes: Reset printer only.
        jmp     NEAR    @@Return        ; No : Return to original program.
@@Init:
        call    BIOS_PrinterInit
        jmp     NEAR    @@Return

@@Reset:
        call    PrinterReset
        jmp     NEAR    @@Return
ENDP    Int09
;...............................................................

;===============================================================
; Interrupt:    =*=Int13=*=
;---------------------------------------------------------------
; Task:         Increment bDisk variable when BIOS int 13h 
;               is used for disk access.
; Input:        none
; Output:       none
; Registers:    none
;====================================================================

;=====  VARIABLES ===================================================
szID13          db      'Prinit (c) Simon Helmut Groes',0  ; ID string.
fDisk           db      0               ; 1=disk access, avoid reentry.
LABEL   dwInt13LPtr     DWORD
wInt13Ofs       dw      0       ; Offset address of old 1Ch int.
wInt13Seg       dw      0       ; Segment address of old 1Ch int.
SEG13           =       $ - wInt13Seg
OFS13           =       $ - wInt13Ofs
;--------------------------------------------------------------------
PROC    Int13

        cmp     [BYTE cs:fDisk], 0      ; Program already running ?
        jne     @@Default               ; Yes: Execute only original int 13.
        inc     [BYTE cs:fDisk]         ; No : Set [fDisk] to 1.
        
        pushf                           ; Simulate 'INT' instruction.
        call    [DWORD cs:dwInt13LPtr]  ; Call original int 13h.

        dec     [BYTE cs:fDisk]         ; Reset [fDisk] flag to 0.
        
        retf    2                       ; Return; add sp, 2;
                                        ;  don't restore flags.
@@Default:
        pushf                           ; Simulate 'INT' instruction.
        call    [DWORD cs:dwInt13LPtr]  ; Call original int 13h.
        retf    2                       ; Return; add sp, 2;
ENDP    Int13                           ;  don't restore flags.
;....................................................................

;====================================================================
; Procedure:    =*=BIOS_PrinterInit=*=
;--------------------------------------------------------------------
; Task:         Initialize printer.
; Input:        none
; Output:       none
; Registers:    none
;====================================================================
PROC    BIOS_PrinterInit

        push    ax
        push    dx

        mov     ah, 01h                 ; BIOS func 1: Initialise printer.
        mov     dx, 0                   ; LPT1
        int     17h                     ; ah = status.
        
        pop     dx
        pop     ax
        ret

ENDP    BIOS_PrinterInit
;....................................................................

;====================================================================
; Procedure:    =*=PrinterReset=*=
;--------------------------------------------------------------------
; Task:         Disable printer by sending a 0 reset bit.
; Input:        none
; Output:       none
; Registers:    none
;====================================================================
PROC    PrinterReset

        push    ax
        push    dx
        push    es

        mov     ax, BIOS_DATASEG
        mov     es, ax
        mov     ax, [es:BIOS_LPT1PORT]
        mov     dx, ax
        inc     dx
        inc     dx                      ; dx = Control Register.
        in      al, dx                  ; al = Printer control reg contents.
        and     al, NOT PC_INIT         ; Zero Reset bit.
        out     dx, al                  ; Reset printer.

        pop     es
        pop     dx
        pop     ax
        ret
ENDP    PrinterReset
;....................................................................

;====================================================================
; Procedure:    =*=ReadKbPort=*=
;--------------------------------------------------------------------
; Task:         Read keyboard port 60h.
; Input:        none
; Output:       al = contents of port. ah = 0.
; Registers:    ax changed.
;====================================================================
PROC    ReadKbPort

        push    dx
        xor     ax, ax
        mov     dx, 60h                 ; Keyboard port address.
        in      al, dx
        pop     dx
        ret
ENDP    ReadKbPort
;....................................................................

;====================================================================
; Procedure:    =*=BIOS_ReadChar=*=
;--------------------------------------------------------------------
; Task:         Read keyboard buffer using BIOS.
; Input:        none
; Output:       ax=0, no keypress in buffer.
;               al=0, extended keypress, ah=keycode
;               al=ASCII code, ah=scancode.
; Registers:    ax changed.
;====================================================================
PROC    BIOS_ReadChar

;-----  First check if keystroke in buffer.
        xor     ah, ah
        inc     ah                      ; ah = function 1, Check kb-buffer.
        int     16h                     ; BIOS kb-services.
        jz      @@NoStroke              ; No keystroke in buffer.
        xor     ax, ax                  ; ah = func 0. Read/update kb-buffer.
        int     16h                     ; al=ASCII code & ah = scan code,
                                        ;  or al=0, ah=extended code.
@@Return:
        ret
@@NoStroke:
        xor     ax, ax
        jmp     NEAR    @@Return
ENDP    BIOS_ReadChar
;....................................................................

;====================================================================
; Procedure:    =*=BIOS_GetKeyState=*=
;--------------------------------------------------------------------
; Task:         Get state of specified key.
; Input:        none
; Output:       al = status bits.
;====================================================================
PROC    BIOS_GetKeyState

        mov     ah, 2           ; Function 2: Get key state.
        int     16h             ; al = key states
        xor     ah, ah          ; ah = 0.
        ret
ENDP    BIOS_GetKeyState
;....................................................................
;********************* TSR END **************************************
TsrEnd:

;********************************************************************
;               INSTALL PROGRAM
;********************************************************************
;-----  DATASEG
szInstalled     db      CR,LF,'PRINIT installed.',0
szAlreadyIn     db      CR,LF,'PRINIT already installed.',0
szNotInstalled  db      CR,LF,'PRINIT not installed.',0
szError7        db      CR,LF,'Memory control blocks destroyed.',0
szError8        db      CR,LF,'Incorrect memory address.',0
szBadArgs       db      CR,LF,'Invalid parameter(s).',0

Begin:
;-----  First check for 'u' parameter.
        call    ARG_ParseArgs           ; ax -> command buffer.
        mov     di, ax                  ; di -> command buffer.
        call    ARG_GetArgNumber        ; Any args ?
        or      ax, ax                  ; No args ?
        jz      Install                 ; No : Install Prinit.
        cmp     ax, 1                   ; More than 1 arg ?
        ja      ArgError
        call    ARG_IsSwitch            ; cf=0, switch; cf=1, not.
        jc      ArgError
        inc     di
        cmp     [BYTE di], 'u'          ; Is switch '/u' ?
        jne     ArgError                ; No : Error.
        jmp     NEAR    UnInstall       ; Yes: Attempt uninstall.

Install:
;-----  Save old interrupt 09h & 13h address.
        push    es
        mov     ax, 3509h               ; Function 35h - Get int vector.
        int     21h                     ; es=seg, bx=ofs
        mov     [cs:wInt09Ofs], bx
        mov     [cs:wInt09Seg], es
        mov     ax, 3513h
        int     21h
        mov     [cs:wInt13Ofs], bx
        mov     [cs:wInt13Seg], es
        pop     es

;-----  Check that Prinit interrupt isn't already installed.
        push    es
        mov     es, [cs:wInt09Seg]
        mov     di, [cs:wInt09Ofs]
        sub     di, ID09                ; di -> ID string.
        mov     si, OFFSET szID09       ; si -> Control ID string.
        call    CHRSTR_StrCmp           ; ax=0, s1=s2, ax=1, s1!=s2.
        pop     es
        or      ax, ax                  ; Strings identical ?
        jne     PrinitInstall           ; No : Install.

;-----  Exit Program: TSR not installed.
        mov     di, OFFSET szAlreadyIn  ; 'PRINIT already installed.'
        call    DOSIO_PutString

UninstTerminate:
        mov     ah, 04Ch                ; Terminate program. al = 1.
        xor     al, al
        inc     al                      ; Return with errcode = 1.
        int     21h

ArgError:
        mov     di, OFFSET szBadArgs    ; 'Invalid parameter(s).'
        call    DOSIO_PutString
        jmp     NEAR    UninstTerminate

PrinitInstall:
;-----  Set new interrupt addresses.
        mov     ax, 2509h               ; Function 25h - Set int vector.
        mov     dx, OFFSET Int09
        int     21h                     ; New int vector set.

        mov     ax, 2513h
        mov     dx, OFFSET Int13
        int     21h

;-----  Free Environment Block:
        mov     es, [cs:ENVBLOCK]       ; Env block seg addr.
        mov     ah, 49h                 ; Free memory block.
        int     21h

;-----  Display that TSR is installed.
        mov     di, OFFSET szInstalled
        call    DOSIO_PutString

;-----  Determine size of TSR in paragraphs.
        mov     dx, OFFSET TsrEnd
        add     dx, 15                  ; Round up.
        mov     cl, 4
        shr     dx, cl                  ; Divide by 16.

;-----  Exit program, leave TSR resident.
        mov     ah, 031h                ; Function 31h - Terminate stay res.
        mov     al, 0                   ; Return with exit code.
        int     21h                     ; Program terminated, TSR resident.

UnInstall:
;-----  Get current int 9 vector:        
        push    es
        mov     ax, 3509h               ; Function 35h - Get int vector.
        int     21h                     ; es=seg, bx=ofs
        mov     [cs:wInt09Ofs], bx
        mov     [cs:wInt09Seg], es
        
;-----  Check that Prinit is installed.
        mov     di, bx
        sub     di, ID09                ; di -> Test ID string
        mov     si, OFFSET szID09       ; si -> Control ID string.
        call    CHRSTR_StrCmp           ; ax=0, s1=s2, ax=1, s1!=s2.
        pop     es
        or      ax, ax
        je      Installed
        jmp     NEAR    NotInstalled

Installed:
;-----  Get current int 13 vector:
        push    es
        mov     ax, 3513h               ; Function 35h - Get int vector.
        int     21h                     ; es=seg, bx=ofs
        mov     [cs:wInt13Ofs], bx
        mov     [cs:wInt13Seg], es
        pop     es

;-----  Restore old interrupts:
;-----  Restore int 9:
        push    es
        mov     es, [cs:wInt09Seg]      ; es = Current int 9 seg addr.
        mov     di, [cs:wInt09Ofs]      ; di = Current int 9 ofs addr.
        push    di                      ; Save di.
        sub     di, OFS09               ; di -> Org int 9 ofs addr.
        mov     ax, [WORD es:di]        ; [es:di] -> ax
        mov     dx, ax                  ; dx = Org int 9 ofs addr.
        pop     di                      ; Restore di (=Current int 9 ofs).
        sub     di, SEG09               ; di -> Org int 9 seg addr.
        mov     ax, [WORD es:di]        ; ax = Org int 9 seg addr.
        push    ds
        mov     ds, ax                  ; ds = Org int 9 seg addr.
        mov     ax, 2509h               ; Func 25h - Set int vector.
        int     21h                     ; Org int 9 vector restored.
        pop     ds
        pop     es


;-----  Restore int 13:
        push    es
        mov     es, [cs:wInt13Seg]
        mov     di, [cs:wInt13Ofs]      ; di = Current int 13 Ofs addr.
        push    di                      ; Save di.
        sub     di, OFS13               ; di -> Original int 13 vector ofs.
        mov     ax, [WORD es:di]        ; ax = Org int 13 vector ofs.
        mov     dx, ax                  ; dx = Org int 13 vector ofs.
        pop     di                      ; Restore di (= Current int 13 ofs).
        sub     di, SEG13               ; di -> Original int 13 vector seg.
        mov     ax, [WORD es:di]        ; ax = Org int 13 seg.
        push    ds                      ; Save ds.
        mov     ds, ax                  ; ds = Org int 13 seg
        mov     ax, 2513h               ; ah = Func 25 - Set int vector.
        int     21h                     ; Org int 13 restored.
        pop     ds
        pop     es

;-----  Free memory occupied by TSR:
        push    es
        mov     es, [cs:wInt09Seg]      ; es = seg addr of memory block.
        mov     ah, 49h                 ; Func 49h - Free memory block.
        int     21h                     ; Memory occupied by TSR freed.
        pop     es
        jc      UninstError
        jmp     NEAR    UninstTerminate

UninstError:
        call    ERRCODE_GetErrMsg       ; ax -> Error message.
        mov     di, ax
        call    DOSIO_PutString
        jmp     NEAR    UninstTerminate

NotInstalled:
        mov     di, OFFSET szNotInstalled
        call    DOSIO_PutString
        jmp     NEAR    UninstTerminate
;***************************************************************

;===============================================================
; Procedure:    =*=CHRSTR_StrLength=*=
;---------------------------------------------------------------
; Task:         Return length of string.
; Input:        (ds:)di = address of string.
; Output:       Length of string in ax.
; Registers:    ax changed
;===============================================================
PROC    CHRSTR_StrLength
        push    cx
        push    di
        push    es

        push    ds
        pop     es              ; es = ds
        mov     al, 0           ; End of string marker.
        mov     cx, 0FFFFh      ; cl <- maximum value (FFFFh = -1).
        cld                     ; Auto increment di
        or      cx, cx          ; Reset zero flag
        repnz   scasb           ; cmp di with al, inc di, dec cl
        neg     cx              ; cx = 1 + stringLength + NULL
        dec     cx              ; cx = stringLength + NULL
        dec     cx              ; cx = stringLength
        mov     ax, cx          ; Return string length in ax.
        
        pop     es
        pop     di
        pop     cx
        ret                     ; Return to caller
ENDP    CHRSTR_StrLength
;...............................................................

;====================================================================
; Procedure:    =*=CHRSTR_StrCmp=*=
;--------------------------------------------------------------------
; Task:         Compare 2 strings to see if they are the same.
; Input:        si = s1, di = s2.
; Output:       ax = 0 - Strings are identical.
;               ax = 1 - Strings are not identical.
; Registers:    ax changed.
; Note:         This procedure is case-sensitive.
;               s1 is the control string, s2 is the test string.
;====================================================================
PROC    CHRSTR_StrCmp

        push    cx
        push    di
        push    si

        xchg    si, di                  ; Xchg for next instruction.
        call    CHRSTR_StrLength        ; ax = strlen - '\0'.
        xchg    si, di                  ; Restore.
        mov     cx, ax
        cld                             ; Auto-inc di & si.
        repe    cmpsb                   ; Cmp s1 & s2.
        xor     ax, ax                  ; ax = 0.
        or      cx, cx                  ; cx = 0 (s1=s2) ?
        jz      @@Return                ; ax = 0, end proc.
        inc     ax                      ; ax = 1, end proc.

@@Return:
        pop     si
        pop     di
        pop     cx
        ret
ENDP    CHRSTR_StrCmp
;....................................................................

;===============================================================
; Procedure:    =*=DOSIO_PutString=*=
;---------------------------------------------------------------
; Task:         Write string to screen using DOS function 40h of int 21h.
; Input:        di = address of string.
; Output:       String written to screen (handle 1: std output device).
; Registers:    none
;===============================================================
PROC    DOSIO_PutString

        push    ax
        push    bx
        push    cx
        push    dx
        push    di

        call    CHRSTR_StrLength        ; Length of string in ax.
        mov     cx, ax                  ; Length of string in cx.
        jcxz    @@Return                ; End procedure if no string.
        mov     dx, di                  ; Offset should be in dx.
        mov     bx, 1                   ; Handle #1 (screen).
        mov     ah, 40h                 ; Function 40 in ah.
        int     21h                     ; Call int 21h.
@@Return:
        pop     di
        pop     dx
        pop     cx
        pop     bx
        pop     ax
        ret
ENDP    DOSIO_PutString
;...............................................................

        END     Start           ; End of program / entry point.

